/* Copyright (c) Colorado School of Mines, 2011.*/
/* All rights reserved.                       */

/* Copyright (c) Colorado School of Mines, 1997.*/
/* All rights reserved.                       */

#include "par.h"

/*********************** self documentation **********************/
/*************************************************************************
This is file VND.h.  It also provides the documentation for the VND routines.
If you use this package of routines outside of SU, you may wish to 
operate independently of the SU defines and replace the include to par.h to by

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/errno.h>
#define	MIN(x,y) ((x) < (y) ? (x) : (y))

Define local file VND_HOME with list of default file systems for VND space
   line 1 of file: 	integer giving number of file systems	
   successive lines: 	directory path to each file system, one per line
*************************************************************************/
/**************** end self doc ********************************/

#define VND_HOME ".VND"

/**************************************************************************
The VND routines are intended for avoiding the need for virtual
memory when prototyping large out-of-core problems where 
multidimensional matrix transposes are required.  The data are 
stored in a multidimensional block matrix transpose format where 
the dimension of the matrix is user defined.  The data are accessed 
by the user as vectors for any given dimension with all the other 
dimensions fixed or defined as a key.  The vector elements all have 
the same user defined number of bytes per element (or node).  The user
specifies how much memory he would like to allocate for buffering
VND I/O operations.  The VND routines break up the multidimensional
matrix into chunks small enough to fit within this memory limitation
if at all possible.  If not, the routines use the smallest amount
of memory that they can.  If the data fits in memory, then it
is put into memory and no I/O operations are required except to 
save on disk.

A dimension over which the data need not be transposed can be defined as the
panel dimension.  If panels are used, then the file structure becomes
that of a sequence of block matrix transpose structures, one for 
each panel.  Normally, only one panel is stored in memory at a time.
The user has the option to specify 2 or more simultaneously open
panels at the expense of using more memory.  The variable "iopanel"
ranges from 0 to the number of open panels minus 1.  The variable
"ipanel" ranges from 0 to the number of panels in the VND files
minus 1.  Two calls to VNDrw with different values of "iopanel"
should not have the same value of "ipanel" unless both are only reading.

There is a preferred dimension for Input/Output efficiency.  That is
the first dimension (idim=0, all of the indicies are zero based).  In
the first dimension, a slice through the block matrix structure can
be read or written in one request.  In all all other dimensions, reading
or writing a slice requires multiple I/O requests.  Therefore, it
is best to choose the first dimension for the dimension with the
most I/O activity.

Real-to-Complex FFT's increase the number of bytes per node and
decrease the number of nodes in a give dimension by a factor
of 2.  To allow this option in dimension 0, the number of nodes
per block in dimension 0  is forced to be even.  This allows
one  to modify the number of bytes per node and number of nodes
in the first dimension to change the data from real to complex
should that be necessary.  This option is not possible for
other than the first dimension.  Clearly, which dimension is 
chosen to be the first is arbitrary other than for considerations
of Input/Output efficiency and this option.

For some applications, it may be desirable to have two dimensions
of size NA and NB in memory at a time to allow rapid changes in
access direction.  Define a single VND dimension with
size NA*NB and access values using the parameters defining starting
element and element increment in VNDrw.  This option keeps user
buffer sizes small and minimizes I/O overhead for some problems.
For implementing the Fowler DMO algorithm, one would use
temporal frequency and velocity together as a function of CMP
position in a single dimension.  This would allow the VND
routines to avoid swapping block matrix slices during the initial 
write of the VND files as the input CMP gathers are stacked for
a range of velocities, Fourier transformed from time to frequency,
and written out to the VND file.  
 
The Unix operating system typically has a 2 gigabyte limit for each 
file system.  When this limit is too small, the VND routines can spread 
the data out over multiple file systems as defined by the user.  Only 
one physical file is actually open at one time to avoid the Unix limitations
on the number of open files allowed.  The code is written in ANSI C and 
should not be limited to the Unix operating system.

The call to VNDop with mode equal to 2 can take a long time as the
first call initializes the file structure to have all zero bytes.
This may involve a significant amount of I/O for large VND files.


The VND user C interface consists of 4 routines:

VND *VNDop(int mode, long lwmax, int ndim, long *N, long npanels, 
		long nbytes, char *file, int ndir, char**dir, int noppanels)
	
	Open block matrix file structure.

	VND *VNDop	returns as pointer to VND I/O control structure
			or NULL if failed
 
	int mode	0 for read only
			1 for read and write (existing file)
			2 for write and read (create new file)
			3 for don't open, just compute buffer sizes 
			and fill out structure
	long lwmax	Maximum memory buffer size in bytes to use
	int ndim	Number of block matrix transpose dimensions
	long N[ndim]	List of number of nodes for each dimension
	long npanels	Number of panels of block matrix structure
	long nbytes	Number of bytes per node
	char *file	File name from which other file names will be 
			constructed.  The actual files used will have
			.VND0, .VND1, ... extentions.  This name
			should not have such an extention.
	int ndir	Number of file directories available for holding data.
			Set ndir=0 to use one file in current directory.
			Set ndir=5 to spread data out over 5 files of
			roughly equal size in the directories dir[j].
			Set ndir=-1 to look at .VND file for directories
			to use.  File format is "number of directories"
			on first line and then one line per directory name.
	char **dir	dir[j] = name of file directory for storing data.
			If dir[0]="/users/junk" and file="file", then the 
			first physical file will be "/users/junk/file.VND0"
	int noppanels	Max number of panels to be open at one time


	VNDop returns a value of NULL if open failed.


void VNDrw(char iop, int iopanel, VND *vnd, int idim, long *key, long ipanel, 
		char *values, long start, long incr, long nvalues,
		int msgnum, char *msg)

	Read or write a vector for dimension "idim" 

	char iop	'r' for read; 'w' for write
	int iopanel	Index of current open panel buffer 
			(Ranges from 0 to noppanels-1.  Used to
			allow access to multiple panels simultaneously.
			For most applications, noppanels = 1 and 
			iopanel will always equal 0.)	
	VND *vnd	Structure holding block matrix file information.
	int idim	Current dimension for read or write
                        (idim=0 is first dimension)
	long key[]	List of node positions for each dimension 
			(Value for dimension "idim" needs to be there
			but it's value is irrelevant.)
	long ipanel	Panel number for read or write.
			(ipanel=0 is first panel)
	char values[]	Returns as values read or input as values to be 
			written.
	long start	starting node (0 based, usually equals 0)
	long incr	node increment (usually 1)
	long nvalues	number of values to read or write 
			(usually same as number of values in 
			the specified dimension)
	int msgnum	A message number to be written out if error occurs.
	char *msg	A message to be written out if error occurs.

void VNDcl(VND *vnd, int mode)

	Close VND file structure.

	VND *vnd	Structure holding VND block matrix file information.
	int mode	0 means keep 
			1 means delete

void VNDflush(VND *vnd)

	Flush out VND buffers to file.  

	This may be important in applications where 
	checkpoint/restart capabilities are needed.  
	After VNDflush, the data written by the VNDrw
	routine are safely stored on disk rather than 
	just in a memory buffer where it could be lost in
	the event of a machine failure.

	VND *vnd	VND control structure

When doing real to complex FFT's, the following routines may
be helpful.

void VNDr2c(VND *vnd)

	Convert dimension 0 from real to complex by scaling
	the number of bytes/node by 2 and halving the
	number of nodes.  (Note: only works for dimension 0)

void VNDc2r(VND *vnd)

	Convert dimension 0 from complex to real by dividing
	the number of bytes/node by 2 and doubling the
	number of nodes.  (Note: only works for dimension 0)

When working in only 2 dimensions, the following routines may
be convenient as they have fewer parameters and don't require
the user to allocate arrays for N or key.  These routines
call the VND routines and are completely compatible.  A FORTRAN
equivalent routine exists with the same calling sequence for
all but the V2Dop routine.  The open routine assumes that
a single VND file can be opened in the current directory.

VND * V2Dop(int mode, long lwmax, long nbytes, char *file,long n0, long n1)

	Open 2-D VND file structure.

	int mode	0 for read only
			1 for read and write (existing file)
			2 for write and read (create new file)
			3 for don't open, just compute buffer sizes 
			and fill out structure
	long lwmax	Maximum memory buffer size in bytes to use
	long nbytes	Number of bytes per node
	char *file	File name from which VND file name will be 
			constructed.  The actual file used will have
			.VND0 extention.  This name
			should not have such an extention.
	long n0		number of nodes in dimension 0
	long n1		number of nodes in dimension 1

void V2Dr0(VND *vnd,long key,char *buf,int msgnum)	read dimension 0
void V2Dr1(VND *vnd,long key,char *buf,int msgnum)	read dimension 1
void V2Dw0(VND *vnd,long key,char *buf,int msgnum)	write dimension 0
void V2Dw1(VND *vnd,long key,char *buf,int msgnum)	write dimension 1

Memory management within the VND routines is based upon
the following routines.  These routines, when used together,
provide automatic error checking at the end of a job
during the VNDfree() step for any buffer overruns telling
the user if an overrun occured at the beginning of the
buffer or at the end.  To ensure that memory overruns
have not occured, a few extra bytes of information must
be stored for each buffer that is not available to the
user program.  Therefore, memory allocated
by VNDemalloc() must be freed by VNDfree() rather than
the standard free() routine.  It is sometimes convenient
to find out the total amount of currently allocated
VND memory.  A routine exists for that purpose.  At the
end of a job, you can ensure that all allocated VND 
buffers have been freed and checked for overruns by
seeing if the total allocated memory at that point
is zero.

void *VNDemalloc(size_t n,char *msg)

	Allocate memory and return a pointer.  
	Give error message and abort if fail.

	size_t n	number of bytes of memory to allocate
	char *msg	error message to print if abort

void VNDmemchk( void *p , char * msg)			

	check for memory overrun and abort if error found
	(can distinguish between overrun at beginning of
	allocated buffer versus overrun at end of allocated
	buffer)

	void *p		pointer returned from VNDemalloc
	char *msg	error message to print out if error found

long VNDtotalmem( void )

	return total current memory allocated by VNDemalloc
	in characters

void VNDfree( void *p, char *msg)

	free memory allocated by VNDemalloc
	(also calls VNDmemchk to check for memory 
	overflows during job)

	void *p		pointer returned from VNDemalloc
	char *msg	error message to print out if error found

	
The following routine returns a unique temporary file
root name for VND files. VNDtempname() 
returns a name made up from the input root string, the
process id, and a unique integer.  This is the only VND
routine calling a UNIX specific function or using the
header files <sys/types.h> or <sys/errno.h>.  The other
routines should be portable to non-Unix systems.

char *VNDtempname(char *root)

Memory associated with a VNDtempname() character string
must be freed using VNDfree() rather than free().

A short test program might check the VND_HOME file for
a .VND file for file systems over which to spread the data.
This has been accomplished by coding "ndir=-1" in VNDop.
A 3 dimensional 50x50x5 block matrix system is set up over
1 panel. The user wishes to use a memory buffer of 1000 
integers. The data are initialized to equal the first dimension
index and written out in vectors along the first dimension (idim=0).
Then, the data are then read from the file along vectors in 
the second dimension (idim=1) and dumped to check and see if 
they are correct.  Note that it would have required memory the
size of 12500 integers to hold the entire 3-D matrix in memory.

#include "VND.h"
int main()
{
int i;
int j;      
int k;      
static long N[3]={50,50,5};
long key[3];
long buf[50];  
VND *vnd;
char **dir;

fprintf(stdout,"Open the VND file and initialize values to 0\n");
vnd=VNDop(2,9999,3,N,1,sizeof(long),"george",-1,dir,1);
VNDdump(vnd,stdout);

fprintf(stdout,"Write out data in first dimension slices\n");
fprintf(stdout,
   "Values = i+j*100+k*10000 where i,j,k are the keys\n");
fprintf(stdout,"in the first, second, and third dimensions\n");
for(k=0;k<N[2];k++) {
   key[2]=k;
   for(j=0;j<N[1];j++) {
	key[1]=j;
        for(i=0;i<N[0];i++) buf[i]=i + j*100 + k*10000;
	VNDrw('w',0,vnd,0,key,0,(char *)buf,0,1,N[0],
		1,"writing 1st dimension (0)");
	}
   }

fprintf(stdout,"\nRead slices along second dimension with k=1 \n");
fprintf(stdout,"Values should equal i+j*100+k*10000\n");
fprintf(stdout,"so for this test, the values in buf should\n");
fprintf(stdout,"increment by 100.  Each line should increment\n");
fprintf(stdout,"by 1.  Since k=1, the starting value is 10000.\n");
key[2]=1;
for(i=0;i<N[0];i++) {
	key[0]=i;
	VNDrw('r',0,vnd,1,key,0,(char *)buf,0,1,N[1],
		2,"reading 2nd dimension (1)");
	printf(" First Dimension = %d  Values 0-4 = %ld %ld %ld %ld %ld\n",
		i,buf[0],buf[1],buf[2],buf[3],buf[4]);		
	}

fprintf(stdout,"\nRead data slices in first dimension\n");
fprintf(stdout,"so that values on a line increment by 1\n");
key[2]=0;
fprintf(stdout,"Third dimension is fixed with k=%ld\n",key[2]);
fprintf(stdout,"Second dimension varies in reverse order\n");
fprintf(stdout,"so lines will decrement by 100.\n");
for(j=N[1]-1;j>=0;j--) {
	key[1]=j;
	VNDrw('r',0,vnd,0,key,0,(char *)buf,0,1,N[0],
		3,"reading 1st dimension in reverse order");
	printf(" Second Dimension = %d  Values 0-4 = %ld %ld %ld %ld %ld\n",
		j,buf[0],buf[1],buf[2],buf[3],buf[4]);		
	}


fprintf(stdout,"\nRead data slices in third dimension\n");
fprintf(stdout,"so values on a line should increment by 10000\n");
key[0]=41;
fprintf(stdout,"First dimension is fixed with i=%ld\n",key[0]);
fprintf(stdout,"Second dimension varies so that successive\n");
fprintf(stdout,"should increment by 100\n");
for(j=0;j<N[1];j++) {
	key[1]=j;
	VNDrw('r',0,vnd,2,key,0,(char *)buf,0,1,N[2],
		4,"reading 3rd dimension");
	printf(" Second Dimension = %d  Values 0-4 = %ld %ld %ld %ld %ld\n",
		j,buf[0],buf[1],buf[2],buf[3],buf[4]);		
	}

VNDcl(vnd,1);
return (0);
}


Utility routines used internally by VNDop, VNDrw, and VNDcl:

int VNDGetDimensions(VND *vnd, long lwmax, int ich)
int VNDrwslice(VND *vnd, char *w, char mode, int idim, long islice, long ipanel)
int VNDseek(VND *vnd, long ib)
void VNDdump(VND *vnd,FILE *fp)

int VNDGetDimensions(VND *vnd, long lwmax, int ich)

	Set dimensions for block matrix transpose 
	in VND control structure.

	Initialize:
		vnd.NumBlocksPerPanel
		vnd.NumBytesMemBuf
		vnd.NumBytesPerBlock
		vnd.NumBytes1stDim
		vnd.NumNodesPerBlock
		vnd.NNodesPerBlock[k]   for k=0 to NumDim-1 
		vnd.NBlocks[k]
		vnd.ByteIncr[k]
		vnd.BlockIncr[k]


	VND *vnd 	VND control structure
			Values required to be set on input
			include vnd.N[k] and vnd.NumBytesPerNode
	long lwmax	number of characters of memory available	
	int ich 	1 if force even number of first dimension
			nodes per block (useful to allow
			real-to-complex ffts and make it
			possible to change number of bytes
			per node at a later point)

	return	-1 if all fits in memory
		0 if successful in partitioning to fit in memory
		1 if failed to partition so fits in memory


int VNDrwslice(VND *vnd,int iopanel, char mode, int idim, 
			long iblock, long ipanel)

	Read or write a slice for dimension idim into or from memory

	VND *vnd	pointer to VND control structure
	int iopanel	open panel counter (0 based. Always
			equals 0 if only one panel open at a time.)
	char mode	'r' for read; 'w' for write.
	int idim	Dimension of slice to be read or written.
			(0 based.  First dimension is idim=0.)
	long iblock	Starting block for block matrix transpose 
			slice associated with dimension idim.
			(Be careful--not sequential 0,1,2,...)
	long ipanel	Panel number (0 based) for desired slice.

	Function returns a value of 0 if successful.

int VNDseek(VND *vnd, long ib)

	Seek the desired block in VND file structure.
	(a) Go to the correct physical file, reopen if necessary
	(b) Seek the correct starting byte within the file

	VND *vnd	pointer to VND control structure
	long ib		desired block number (0 based,
			covers all panels and all files)

	The function returns 0 if successful, Nonzero if fail.

void VNDdump(VND *vnd,FILE *fp)

	Dump out values stored in VND structure to file.

	VND *vnd	pointer to VND control structure
	FILE *fp	FILE pointer associated with output file


void VNDerr(char *msg);

	Abort and print message.
	char *msg	message to print prior to abort.

***************************************************************************/

/***************************************************************************
This is the Fortran user interface to the VND routines.


VNDOP(vnd, mode, lwmax, ndim, N, npanels, 
		nbytes, file, ndir, dir, noppanels)
	
	Open block matrix file structure.

	int vnd		returns as pointer to filled out I/O 
			control structure 
	integer mode	0 for read only
			1 for read and write (existing file)
			2 for write and read (create new file)
			3 for don't open, just compute buffer sizes 
			and fill out structure
	integer lwmax	Maximum memory buffer size in nodes to use
	integer ndim	Number of block matrix transpose dimensions
	integer N[ndim]	List of number of nodes for each dimension
	integer npanels	Number of panels of block matrix structure
	integer nbytes	Number of bytes per node
	character*80 file File name from which other file names will be 
			constructed.  The actual files used will have
			.VND0, .VND1, ... extentions.  This name
			should not have such an extention.
	integer ndir	Number of file directories available for holding data.
			Set ndir=0 to use one file in current directory.
			Set ndir=-1 to look at .VND file for directories
			to use.  File format is "number of directories"
			on first line and then one line per directory name.
	character*80 dir[] list of directories for VND files
	int noppanels	Max number of panels to be open at one time


VNDRW(iop, iopanel, vnd, idim, key, ipanel, 
		values, start, incr, nvalues, msgnum)

	Read or write a vector for dimension "idim" 

	character*1 iop	"r" for read; "w" for write
	integer iopanel	Index of current open panel buffer 
			(Ranges from 1 to noppanels.  Used to
			allow access to multiple panels simultaneously.
			For most applications, noppanels = 1 and 
			iopanel will always equal 1.)	
	integer vnd	Structure holding block matrix file information.
	integer idim	Current dimension for read or write
			(idim=1 for first dimension.)
	integer key[]	List of node positions for each dimension 
			(Value for dimension "idim" needs to be there
			but it's value is irrelevant.)
	integer ipanel	Panel number for read or write.
			(ipanel=1 for first panel.)
	integer values[]Returns as values read or input as values to be 
			written.
	integer start	starting node (First node is 1, usually equals 1)
	integer incr	node increment (usually 1)
	integer nvalues	number of values to read or write 
			(usually same as number of values in 
			the specified dimension)
	integer msgnum	A message number to be written out if error occurs.

VNDCL(vnd, mode)

	Close VND file structure.

	integer vnd	Structure holding VND block matrix file information.
	integer mode	0 means keep 
			1 means delete


VNDFLUSH(vnd)

	Flush out VND buffers to file.  

	This may be important in applications where 
	checkpoint/restart capabilities are needed.  
	After VNDFLUSH, the data written by the VNDRW
	routine are safely stored on disk rather than 
	just in a memory buffer where it could be lost in
	the event of a machine failure.

	integer vnd	VND control structure

An example Fortran code similar to the C code example above
appears below.  On the Cray (or Convex in PD8 mode), be sure
to specify the bytes per node to be 8 instead of 4 in the
the VNDOP call.

	PROGRAM EXAMPLE
	INTEGER VND, N(3), KEY(3), BUF(50), I, J, K
	CHARACTER*80 FILE
	CHARACTER*80 DIR(2)
	FILE='GEORGE'
	N(1)=50
	N(2)=50
	N(3)=5
	OPEN(10,FILE='LISTING')
	CALL VNDOP(VND,2,1000,3,N,1,4,FILE,-1,DIR,1)

	WRITE(10,*) ' '
	WRITE(10,*) ' WRITE OUT THE DATA IN THE FIRST DIMENSION'
	WRITE(10,*) ' DATA VALUES = I + J*100 + K*10000'
	WRITE(10,*) ' WHERE I=1ST DIM, J=2ND DIM, AND K=3RD DIM'
	DO 300 K=1,N(3)
	   KEY(3)=K
	   DO 200 J=1,N(2)
		DO 100 I=1,N(1)
  100		BUF(I) = I + 100*J + 10000*K
		KEY(2)=J
		CALL VNDRW('W',1,VND,1,KEY,1,BUF,1,1,N(1),1)
  200      CONTINUE
  300	CONTINUE

	WRITE(10,*) ' '
	WRITE(10,*) ' READ THE DATA ALONG DIMENSION 2 '
	WRITE(10,*) ' THE 3RD DIMENSION IS HELD FIXED WITH K=2'
	WRITE(10,*) ' VALUES ON A LINE SHOULD INCREMENT BY 100'
	WRITE(10,*) ' VALUES SHOULD INCREMENT BY 1 FROM LINE TO LINE'
	KEY(3)=2
	DO 400 I=1,N(1)
		KEY(1)=I
		CALL VNDRW('R',1,VND,2,KEY,1,BUF,1,1,N(2),2)
		WRITE(10,*) ' I = ',I,' BUF(1-5)=',(BUF(J),J=1,5)
  400	CONTINUE

	WRITE(10,*) ' '
	WRITE(10,*) ' READ THE DATA ALONG DIMENSION 1'
	WRITE(10,*) ' THE 3RD DIMENSION IS HELD FIXED WITH K=1'
	WRITE(10,*) ' VALUES ON A LINE SHOULD INCREMENT BY 1'
	WRITE(10,*) ' VALUES SHOULD DECREMENT BY 100'
	KEY(3)=1
	DO 500 J=1,N(2)
		KEY(2)=N(2)+1-J
		CALL VNDRW('R',1,VND,1,KEY,1,BUF,1,1,N(1),3)
		WRITE(10,*) ' J = ',KEY(2),' BUF(1-5)=',(BUF(I),I=1,5)
  500	CONTINUE

	WRITE(10,*) ' '
	WRITE(10,*) ' READ THE DATA ALONG DIMENSION 3'
	WRITE(10,*) ' THE 1ST DIMENSION IS HELD FIXED WITH I=41 '
	WRITE(10,*) ' VALUES ON A LINE SHOULD INCREMENT BY 10000'
	WRITE(10,*) ' VALUES SHOULD INCREMENT BY 100 FROM LINE TO LINE'
	KEY(1)=41
	DO 600 J=1,N(2)
		KEY(2)=J
		CALL VNDRW('R',1,VND,3,KEY,1,BUF,1,1,N(3),4)
		WRITE(10,*) ' J = ',J,' BUF(1-5)=',(BUF(K),K=1,5)
  600	CONTINUE


	CALL VNDCL(VND,1)	
	CLOSE(10)
	STOP
	END

*************************************************************************/

/* MaxIOBuf is the maximum length of an Input/Output operation in bytes */
/* allowed on current system.	If it is small, there may be no 	*/
/* advantage to using a large amount of memory for VND buffering so	*/
/* VND memory sizes are adjusted accordingly.				*/
/* #define MaxIOBuf 32768						*/

/* BytesPerSector equals the number of bytes per physical disk sector.	*/
/* If it is set to 1, then memory and i/o blocks for the VND routines	*/
/* are not rounded to to the next larger sector size so there is not	*/
/* any wasted space and files and memory buffers are as small as 	*/
/* possible.  On some machines, it is important to work in blocks that 	*/
/* are multiples of the sector size for efficiency reasons.  In that 	*/
/* case, use								*/
/*									*/
/* #define BytesPerSector BUFSIZ					*/
/*									*/
/* Note that BUFSIZ is defined in the stdio.h header file.		*/ 


/* Define integer size and routine names for Fortran interface		*/

/* Use this on CONVEX only if want 8 byte word length for */
/* integers or floats.  Usually use the default. */
#ifdef CONVEXPD8
#define FORTINT long long int
#define BytesPerSector BUFSIZ
#define MaxIOBuf BUFSIZ*16

/* The IBM RS6000 workstations assume Fortran routines */
/* are all lower case if you are not using special compiler */
/* options as are standardly done in systems like ProMAX.  */ 
/* For use with ProMAX, don't specify the compiler option */
/* -DIBMRS6000 during your compiliation. */
#elif IBMRS6000
#define FORTINT long
#define vndop_ vndop
#define vndrw_ vndrw
#define vndcl_ vndcl
#define vndflush_ vndflush
#define v2dr2c_ v2dr2c
#define v2dc2r_ v2dc2r
#define v2dr1_ v2dr1
#define v2dr2_ v2dr2
#define v2dw1_ v2dw1
#define v2dw2_ v2dw2
#define BytesPerSector BUFSIZ
#define MaxIOBuf BUFSIZ*16

/*  The CRAY assumes Fortran routines are capitalized.  */
/*  Reals and ints are 8 bytes long.  Very long i/o's are best. */
/*  New documentation suggests there is a better way to */
/*  code Fortran string variables than done here for the Cray.  */
#elif CRAY
#define FORTINT int
#define vndop_ VNDOP
#define vndrw_ VNDRW
#define vndcl_ VNDCL
#define vndflush_ VNDFLUSH
#define v2dr2c_ V2DR2C
#define v2dc2r_ V2DC2R
#define v2dr1_ V2DR1
#define v2dr2_ V2DR2
#define v2dw1_ V2DW1
#define v2dw2_ V2DW2
#define BytesPerSector BUFSIZ
#define MaxIOBuf BUFSIZ*1024 

/*  This works pretty well as the default on most machines. */
#else
#define FORTINT long
#define BytesPerSector BUFSIZ
#define MaxIOBuf BUFSIZ*8

#endif
typedef struct{
	long NumDim;		/* number of active dimensions 			*/
	long NumBytesPerNode;	/* number of bytes per node 			*/
	long NumBlocksPerPanel;	/* number of blocks per panel 			*/
	long NumFiles;		/* number of physical files used 		*/
	long NumPanels;		/* number of panels 				*/
	long NumOpenPanels;	/* number of panels open at one time 		*/
	long NumBytesMemBuf;	/* number of bytes in memory buffer 		*/
	long NumBytesPerBlock;  /* number of bytes per block 			*/
	long NumBytes1stDim;	/* number of bytes for IO in 1st dimension 	*/
	long NumNodesPerBlock;	/* number of nodes per block 			*/
	long NumBlocksPerFile;	/* number of blocks per physical file 		*/
	int CurrentFile;	/* index (zero based) of current open file 	*/
				/* -1 if not open 				*/
	char *Mode;		/* file open mode: "rb", "rb+", or "wb+"	*/
	FILE *FileDescriptor;	/* file descriptor for ANSI standard binary IO 	*/

   				/* The following have "physical file number" as */
				/* the first index 				*/

	int *LenFileNames;	/* length of each file name in characters 	*/
	char **FileNames;	/* list of file names for physical files 	*/
	long *FirstDataByte;	/* first data byte for each physical file 	*/
	long *LenFile;		/* length of each physical file in bytes 	*/

   				/*The following are arrays dimensioned NumDim.	*/
	long *N;		/* number of nodes for each dimension 		*/
	long *NNodesPerBlock;	/* array of nodes per block for each dimension 	*/
	long *NBlocks;		/* number of blocks in each dimension 		*/
	long *ByteIncr;		/* byte increment between values in each 	*/
				/* dimension 					*/
	long *BlockIncr;	/* block increment for each dimension 		*/

  				/* The following are arrays dimensioned 	*/
				/* noppanels, one for each panel to open 	*/
				/* simultaneously.				*/
  
	long *CurrentPanel;	/* current panel in memory (zero based) 	*/
	int *CurrentDimension;	/* current dimension, -1 if none in memory 	*/
	long *CurrentBlock;	/* starting block for current slice in memory 	*/
				/* for current dimension 			*/
	int *Modified;		/* 1 if current memory buffer has been modified */
				/* 0 if current memory not modified and can be	*/
				/*   thrown away without losing any information	*/

	char *w;		/* pointer to internal memory buffer used by VND*/
} VND;

VND *VNDop(int mode, long lwmax, int ndim, long *N, long npanels, 
		long nbytes, char *file, int ndir, char**dir, int noppanels);
void VNDrw(char iop, int iopanel,VND *vnd,int idim, 
		long *key, long ipanel, char *values,
		long start, long incr, long nvalues,
		int msgnum, char *msg);
void VNDcl(VND *vnd, int mode);
void VNDflush(VND *vnd);
char *VNDtempname(char *root);
int VNDGetDimensions(VND *vnd, long lwmax, int ich);
int VNDrwslice(VND *vnd, int iopanel, char mode, int idim, long iblock, long ipanel);
int VNDseek(VND *vnd, long ib);
void VNDdump(VND *vnd,FILE *fp);
void VNDlook(int *w, int lw);
void *VNDemalloc(size_t n,char *msg);
void VNDerr(char *msg);

/* additions 2/94 to the C interface */
void VNDr2c(VND *vnd);
void VNDc2r(VND *vnd);
VND *V2Dop(int mode, long lwmax, long nbytes, char *file, long n0, long n1);
void V2Dr0(VND *vnd,long key,char *buf,int msgnum);
void V2Dr1(VND *vnd,long key,char *buf,int msgnum);
void V2Dw0(VND *vnd,long key,char *buf,int msgnum);
void V2Dw1(VND *vnd,long key,char *buf,int msgnum);

/* additions 2/94 to the Fortran interface */
void vndr2c_(VND *vnd);
void vndc2r_(VND *vnd);
void v2dr1_(VND *vnd,FORTINT key,void *buf,FORTINT msgnum);
void v2dr2_(VND *vnd,FORTINT key,void *buf,FORTINT msgnum);
void v2dw1_(VND *vnd,FORTINT key,void *buf,FORTINT msgnum);
void v2dw2_(VND *vnd,FORTINT key,void *buf,FORTINT msgnum);

/* additions 9/94 to the C interface */
void VNDmemchk( void *p , char * msg);
void VNDfree( void *p, char *msg);
long VNDtotalmem( void );

#ifdef CRAY
void vndop_( VND **vnd, FORTINT *fort_mode, FORTINT *fort_lwmax, 
	FORTINT *fort_ndim, FORTINT *fort_N, FORTINT *fort_npanels,  
	FORTINT *fort_nbytes, char *fort_file, 
	FORTINT *fort_ndir, char *fort_dir, FORTINT *fort_noppanels);
void vndrw_(char *fort_iop, FORTINT *fort_iopanel,VND **vnd,
	FORTINT *fort_idim, FORTINT *fort_key, FORTINT *fort_ipanel,
	void *values, FORTINT *fort_start, FORTINT *fort_incr,
	FORTINT *fort_nvalues, FORTINT *fort_msgnum);
#else
void vndop_( VND **vnd, FORTINT *fort_mode, FORTINT *fort_lwmax, 
	FORTINT *fort_ndim, FORTINT *fort_N, FORTINT *fort_npanels,  
	FORTINT *fort_nbytes, char *fort_file, 
	FORTINT *fort_ndir, char *fort_dir, FORTINT *fort_noppanels, 
	int lfile, int ldir);
void vndrw_(char *fort_iop, FORTINT *fort_iopanel,VND **vnd,
	FORTINT *fort_idim, FORTINT *fort_key, FORTINT *fort_ipanel,
	void *values, FORTINT *fort_start, FORTINT *fort_incr,
	FORTINT *fort_nvalues, FORTINT *fort_msgnum, long liop);
#endif
void vndcl_(VND **vnd, FORTINT *fort_mode);
void vndflush_(VND **vnd);

#ifdef IBMRS6000
void *malloc();
void free();
#endif


